import os
import time


class ProgressBar:
    def __init__(self):

        self.start_time = time.time()

    def set_total(self, total):

        self.total = total

    def report(self, iteration):

        self.print_progress_bar(iteration, self.total)

    def completion_report(self):

        self.print_progress_bar(self.total, self.total, suffix="Completed!")

    def print_progress_bar(self, iteration, total, suffix="", length=50, fill="█"):

        dur = time.time() - self.start_time
        time_str = time.strftime("%H:%M:%S", time.gmtime(dur))

        percent = ("{0:.1f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + "-" * (length - filledLength)
        print(f"\rProgress: |{bar}| {percent}% | {time_str}  {suffix}", end="\r")
        if iteration == total and suffix == "Completed!":
            print(f"\r", end="\n")


def check_pid(pid):
    """Check For the existence of a unix pid."""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True



import time
from collections import OrderedDict
from multiprocessing import Process

import msgpack
import zmq

from .config import *
from .utils import check_pid


class WorkerQueue(object):
    """Worker queue used by the server for load balancing. It uses
    the list data structure for queue. It also keeps track of which
    worker is busy processing which task key. The set of all worker
    address is used by, for instance, kill command.
    """

    def __init__(self):
        self.free_workers = list()
        self.busy_workers = dict()
        self.all_workers = set()

    def ready(self, worker_address):

        if worker_address in self.free_workers:
            self.free_workers.remove(worker_address)

        self.free_workers.append(worker_address)
        self.all_workers.add(worker_address)

    def busy(self, worker_address, task_key):

        self.busy_workers[worker_address] = task_key

    def free(self, worker_address):

        self.busy_workers.pop(worker_address, None)

    def next(self):

        address = self.free_workers.pop(False)
        return address

    def purge(self):
        """Only called when the client detects dead worker processes.
        It purges any record of the dead worker from the queue's state.
        Also, it will retrieve the keys of tasks that those dead workers were
        unable to finish processing in ``unfinished_tasks`` and pass
        them along to the client for further instructions.
        """
        unfinished_tasks = []

        all_workers = self.all_workers.copy()
        for address in all_workers:

            pid = int(address.decode())
            if not check_pid(pid):

                if address in self.free_workers:
                    self.free_workers.remove(address)

                if address in self.busy_workers:
                    task_key = self.busy_workers.pop(address)
                    unfinished_tasks.append(task_key)

                if address in self.all_workers:
                    self.all_workers.remove(address)

        return unfinished_tasks


class ServerProcess(Process):
    """The Proxy Server that the Futures client and worker clients connect to.
    It uses asynchronous messaging queues to communicate with clients. Its main
    functionality is load balancing, which is implemented as a queue of LRU workers
    (refer to ``WorkerQueue``).

    Parameters
    ----------
    client_address : str

    verbose : int
    """

    def __init__(self, client_address, verbose):
        super(ServerProcess, self).__init__()

        self.client_address = client_address
        self.verbose = verbose

    def print(self, s, lvl):

        if self.verbose >= lvl:
            print(s)

    def run(self):

        context = zmq.Context(1)

        # Frontend server socket (to talk to client)
        frontend = context.socket(zmq.ROUTER)
        frontend.bind(SERVER_ENDPOINT)

        # Backend server socket (to talk to workers)
        backend = context.socket(zmq.ROUTER)
        backend.bind(WORKER_ENDPOINT)

        # Poller for workers
        poll_workers = zmq.Poller()
        poll_workers.register(backend, zmq.POLLIN)

        # Poller for both workers and client
        poll_both = zmq.Poller()
        poll_both.register(frontend, zmq.POLLIN)
        poll_both.register(backend, zmq.POLLIN)

        # Queue of LRU workers
        workers = WorkerQueue()

        try:

            # Flag for the ``finally`` clause.
            perform_final = True

            while True:

                # If workers queue is empty, we wait for workers to send the initial
                # WORKER_READY_SIGNAL signal so we can queue the ready workers.
                if len(workers.free_workers) > 0:
                    poller = poll_both
                else:
                    poller = poll_workers

                # Start listening to client/workers (blocking)
                socks = dict(poller.poll())

                # Get message from the backend worker.
                if socks.get(backend) == zmq.POLLIN:

                    # [worker_address, client_address, task_key, task_success_signal, result] or
                    # [worker_address, client_address, task_key, task_failure_signal, error] or
                    # [worker_address, worker_ready_signal]
                    # The DEALER socket prepended the worker address.
                    frames = backend.recv_multipart()

                    self.print(
                        "6. RECEIVED IN SERVER FROM WORKER RAW: {}\n\n".format(frames),
                        3,
                    )

                    # Interrupted.
                    if not frames:
                        break

                    # Strip the prepended worker address.
                    address = frames[0]

                    # The worker reply payload.
                    # [client_address, task_key, task_success_signal, result] or
                    # [client_address, task_key, task_failure_signal, error] or
                    # [worker_ready_signal]
                    reply_payload = frames[1:]

                    if len(reply_payload) == 1:

                        if reply_payload[0] == WORKER_READY_SIGNAL:

                            workers.ready(address)

                        else:

                            print("E: Invalid message from worker: %s" % reply_payload)

                    else:

                        if frames[3] == TASK_SUCCESS_SIGNAL:

                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: {}\n".format(
                                    frames
                                ),
                                2,
                            )
                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: [worker_address, client_address, task_key, task_success_signal, result]\n\n",
                                1,
                            )

                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: {}\n".format(
                                    reply_payload
                                ),
                                2,
                            )
                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: [client_address, task_key, task_success_signal, result]\n\n",
                                1,
                            )

                        if frames[3] == NUMPY_TASK_SUCCESS_SIGNAL:

                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: {}\n".format(
                                    frames
                                ),
                                2,
                            )
                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: [worker_address, client_address, task_key, numpy_task_success_signal, metadata, result]\n\n",
                                1,
                            )

                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: {}\n".format(
                                    reply_payload
                                ),
                                2,
                            )
                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: [client_address, task_key, numpy_task_success_signal, metadata, result]\n\n",
                                1,
                            )

                        if frames[3] == TASK_FAILURE_SIGNAL:

                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: {}\n".format(
                                    frames
                                ),
                                1,
                            )
                            self.print(
                                "6. RECEIVED IN SERVER FROM WORKER: [worker_address, client_address, task_key, task_failure_signal, error]\n\n",
                                1,
                            )

                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: {}\n".format(
                                    reply_payload
                                ),
                                1,
                            )
                            self.print(
                                "7. SENDING FROM SERVER TO CLIENT: [client_address, task_key, task_failure_signal, error]\n\n",
                                1,
                            )

                        # Remove the worker from busy list.
                        workers.free(address)
                        workers.ready(address)

                        # [client_address, task_key, task_success_signal, result] or
                        # [client_address, task_key, task_failure_signal, error]
                        # The ROUTER socket will strip off the client address.
                        frontend.send_multipart(reply_payload)

                # Get message from the frontend client.
                if socks.get(frontend) == zmq.POLLIN:

                    # The DEALER socket prepended the client address.
                    # [client_address, task_key, task_mode_signal, start_method_signal, func_statefulness_signal, func, args]
                    # for normal task request.
                    # [client_address, dummy_task_key, kill_signal]
                    # if ``close`` is invoked from futures client.
                    # [client_address, dummy_task_key, worker_failure_signal]
                    # if any worker process termination is detected from futures client.
                    frames = frontend.recv_multipart()

                    self.print(
                        "2. RECEIVED IN SERVER FROM CLIENT RAW: {}\n\n".format(frames),
                        3,
                    )

                    # Interrupted
                    if not frames:
                        self.print("2. RECEIVED IN SERVER FROM CLIENT: NULL\n\n", 1)
                        break

                    if frames[2] in [
                        NORMAL_TASK_REQUEST_SIGNAL,
                        PANDAS_PARTITION_TASK_REQUEST_SIGNAL,
                        PANDAS_NONPARTITION_TASK_REQUEST_SIGNAL,
                    ]:

                        self.print(
                            "2. RECEIVED IN SERVER FROM CLIENT: {}\n".format(frames), 2
                        )
                        self.print(
                            "2. RECEIVED IN SERVER FROM CLIENT: [client_address, task_key, task_mode_signal, "
                            "start_method_signal, func_statefulness_signal, func, args]\n\n",
                            1,
                        )

                        # Get the address of LRU worker
                        address = workers.next()

                        # Get the task key
                        task_key = msgpack.unpackb(frames[1], raw=False)

                        # Prepend the LRU worker address.
                        # [worker_address, client_address, task_key, task_mode_signal, start_method_signal, func_statefulness_signal, func, args]
                        frames.insert(0, address)

                        self.print(
                            "3. SENDING FROM SERVER TO WORKER: {}\n".format(frames), 2
                        )
                        self.print(
                            "3. SENDING FROM SERVER TO WORKER: [worker_address, client_address, task_key, task_mode_signal, "
                            "start_method_signal, func_statefulness_signal, func, args]\n\n",
                            1,
                        )

                        # The ROUTER will strip off the worker address from the frames and save it in
                        # hash table, associating it with the worker socket.
                        # [worker_address, client_address, task_key, task_mode_signal, start_method_signal, func_statefulness_signal, func, args]
                        backend.send_multipart(frames)

                        workers.busy(address, task_key)

                    elif frames[2] == KILL_SIGNAL:

                        self.print(
                            "*. RECEIVED IN SERVER FROM CLIENT: {}\n".format(frames), 2
                        )
                        self.print(
                            "*. RECEIVED IN SERVER FROM CLIENT: [client_address, dummy_task_key, kill_signal]\n\n",
                            1,
                        )

                        for worker_address in workers.all_workers:

                            msg = [worker_address, KILL_SIGNAL]

                            # The ROUTER will strip off the worker address from the frames and save it in
                            # hash table, associating it with the worker socket.
                            # [worker_address, kill_signal]
                            backend.send_multipart(msg)

                        break

                    elif frames[2] == WORKER_FAILURE_SIGNAL:

                        self.print(
                            "5. RECEIVED IN SERVER FROM CLIENT: {}\n".format(frames), 2
                        )
                        self.print(
                            "5. RECEIVED IN SERVER FROM CLIENT: [client_address, dummy_task_key, worker_failure_signal]\n\n",
                            1,
                        )

                        failed_task_keys = workers.purge()
                        failed_task_keys = msgpack.packb(
                            failed_task_keys, use_bin_type=True
                        )

                        msg = [
                            self.client_address,
                            DUMMY_TASK_KEY,
                            WORKER_FAILURE_SIGNAL,
                            failed_task_keys,
                        ]

                        frontend.send_multipart(msg)

        except KeyboardInterrupt:

            print("Keyboard Interrupted")

            perform_final = True

        except ZMQError:

            perform_final = False

        finally:

            if perform_final:

                poll_both.unregister(frontend)
                poll_both.unregister(backend)
                poll_workers.unregister(backend)

                frontend.setsockopt(zmq.LINGER, 0)
                frontend.close()

                backend.setsockopt(zmq.LINGER, 0)
                backend.close()

                context.term()



REQUEST_TIMEOUT = 2500

#  Paranoid Pirate Protocol constants
WORKER_READY_SIGNAL = b"\x01"  # Signals worker is ready
KILL_SIGNAL = b"\x03"

# Task modes signals
NORMAL_TASK_REQUEST_SIGNAL = b"\x04"
PANDAS_PARTITION_TASK_REQUEST_SIGNAL = b"\x05"
PANDAS_NONPARTITION_TASK_REQUEST_SIGNAL = b"\x06"

# Process start method signals
FORKED_PROCESS_SIGNAL = b"\x07"
SPAWNED_PROCESS_SIGNAL = b"\x08"

# Statefulness method signals
STATEFUL_METHOD_SIGNAL = b"\x10"  # b"\x9" is reserved
STATELESS_METHOD_SIGNAL = b"\x11"

# Task result signals
TASK_SUCCESS_SIGNAL = b"\x12"
TASK_FAILURE_SIGNAL = b"\x13"
WORKER_FAILURE_SIGNAL = b"\x14"
NUMPY_TASK_SUCCESS_SIGNAL = b"\x15"

# Endpoint address
SERVER_ENDPOINT = "tcp://127.0.0.1:5555"
WORKER_ENDPOINT = "tcp://127.0.0.1:5556"

DUMMY_TASK_KEY = b"\x16"

TMP_FILENAME = "MY_tmp_"



import os
import sys
import time
import warnings
from multiprocessing import Process
from random import randint

import cloudpickle
import feather
import msgpack
import numpy as np
import pandas as pd
import zmq
from zmq.error import ZMQError

from .config import *


class TASK_FAILED(UserWarning):
    pass


def worker_socket(context, poller):
    """Helper function that returns a new configured socket
    connected to the Paranoid Pirate queue"""
    identity = str(os.getpid()).encode()
    worker = context.socket(zmq.DEALER)  # DEALER ~ requester
    worker.setsockopt(zmq.IDENTITY, identity)
    poller.register(worker, zmq.POLLIN)
    worker.connect(WORKER_ENDPOINT)
    worker.send(WORKER_READY_SIGNAL)
    return identity, worker


class WorkerProcess(Process):
    """The process for the worker client that connects to the server and carries
    out the requested computations in parallel. The user function can access its
    state if needed. The underscore is for when the user wishes to subclass the
    BaseWorker.

    Parameters
    ----------
    verbose : int
        Available input values:
        - 0: no output
        - 1: show message titles
        - 2: show message contents

    dataframe : Pandas dataframe or None
        Used if ``apply`` method is invoked.
    """

    def __init__(
        self,
        __verbose__,
        __dataframe__=None,
        __forked__=True,
        __mode__="normal",
        __partition__=False,
        *args,
        **kwargs
    ):
        super(WorkerProcess, self).__init__()

        self.__verbose__ = __verbose__

        if __mode__ == "pandas":

            if __forked__:

                self.__dataframe__ = __dataframe__

            else:

                if __partition__:

                    pass

                else:

                    self.__dataframe__ = feather.read_dataframe(__dataframe__)

        else:

            self.__dataframe__ = None

    def print(self, s, lvl):

        if self.__verbose__ >= lvl:
            print(s)

    def run(self):

        context = zmq.Context(1)
        poller = zmq.Poller()

        identity, worker = worker_socket(context, poller)
        self.print("WORKER STARTED: {}".format(identity), 1)

        try:

            # Flag for the ``finally`` clause.
            perform_final = True

            while True:

                socks = dict(poller.poll())

                # Get message from the proxy server.
                if socks.get(worker) == zmq.POLLIN:

                    # The ROUTER socket strips the worker_address.
                    # [client_address, task_key, task_mode_signal, start_method_signal, func_statefulness_signal, func, args] or
                    # for normal task request.
                    # [kill_signal]
                    # if ``close`` is invoked from futures client.
                    frames = worker.recv_multipart()

                    # Interrupted
                    if not frames:
                        self.print(
                            "4. RECEIVED IN WORKER {} FROM SERVER: NULL\n\n".format(
                                identity
                            ),
                            1,
                        )
                        break

                    # Kill signal from the frontend client.
                    if len(frames) == 1 and frames[0] == KILL_SIGNAL:
                        break

                    # Task request from the frontend client.
                    elif frames[2] in [
                        NORMAL_TASK_REQUEST_SIGNAL,
                        PANDAS_PARTITION_TASK_REQUEST_SIGNAL,
                        PANDAS_NONPARTITION_TASK_REQUEST_SIGNAL,
                    ]:

                        self.print(
                            "4. RECEIVED IN WORKER {} FROM SERVER: {}\n".format(
                                identity, frames
                            ),
                            2,
                        )
                        self.print(
                            "4. RECEIVED IN WORKER {} FROM SERVER: [client_address, task_key, task_mode_signal, start_method_signal, "
                            "func_statefulness_signal, func, args]\n\n".format(
                                identity
                            ),
                            1,
                        )

                        # [client_address, task_key]
                        reply_frame_header = frames[:2]
                        # [task_mode_signal, start_method_signal, func_statefulness_signal]
                        task_properties = frames[2:5]
                        # [func, args]
                        task_payload = frames[5:]

                        func_binary = task_payload[0]
                        func = cloudpickle.loads(func_binary)

                        args_binary = task_payload[1]
                        args, kwargs = msgpack.unpackb(args_binary, raw=False)

                        self.print("ARGS: {}\n\n".format(args), 3)
                        self.print("KWARGS: {}\n\n".format(kwargs), 3)

                        # Try to catch user function exception without disturbing the process if possible.
                        try:

                            # Run the user ``func``.
                            if task_properties[0] == NORMAL_TASK_REQUEST_SIGNAL:

                                if task_properties[2] == STATEFUL_METHOD_SIGNAL:

                                    self.print("MODE: NORMAL, STATEFUL", 1)

                                    result = func(self, *args, **kwargs)

                                elif task_properties[2] == STATELESS_METHOD_SIGNAL:

                                    result = func(*args, **kwargs)

                            elif (
                                task_properties[0]
                                == PANDAS_PARTITION_TASK_REQUEST_SIGNAL
                            ):

                                # If we are using ``apply`` method, the ``func`` is required to have the
                                # dataframe as its first positional argument.
                                result = func(self.__dataframe__, *args, **kwargs)

                                if isinstance(result, (pd.DataFrame, pd.Series)):
                                    result = result.values

                            elif (
                                task_properties[0]
                                == PANDAS_NONPARTITION_TASK_REQUEST_SIGNAL
                            ):

                                # If we are using ``apply`` method, the ``func`` is required to have the
                                # dataframe as its first positional argument.
                                result = func(self.__dataframe__, *args, **kwargs)

                                if isinstance(result, (pd.DataFrame, pd.Series)):
                                    result = result.values

                            else:

                                raise ValueError("Invalid task_model_signal.")

                            # Communicate the results back to the server.
                            # If the result is a numpy n-dim array, use memory buffer for messaging.
                            if isinstance(result, (np.ndarray, np.generic)):

                                if not result.flags["C_CONTIGUOUS"]:
                                    result = np.ascontiguousarray(result)

                                metadata = dict(
                                    dtype=str(result.dtype),
                                    shape=result.shape,
                                )
                                metadata = msgpack.packb(metadata, use_bin_type=True)

                                reply_frame = reply_frame_header + [
                                    NUMPY_TASK_SUCCESS_SIGNAL,
                                    metadata,
                                    result,
                                ]

                                self.print(
                                    "5. SENDING FROM WORKER {} TO SERVER: {}\n".format(
                                        identity, reply_frame
                                    ),
                                    2,
                                )
                                self.print(
                                    "5. SENDING FROM WORKER {} TO SERVER: [client_address, task_key, numpy_task_success_signal, "
                                    "metadata, result]\n\n".format(identity),
                                    1,
                                )

                            # If not, use the msgpack to serialize the return data.
                            else:

                                result_binary = msgpack.packb(result, use_bin_type=True)

                                # [client_address, task_key, task_success_signal, result]
                                reply_frame = reply_frame_header + [
                                    TASK_SUCCESS_SIGNAL,
                                    result_binary,
                                ]

                                self.print(
                                    "5. SENDING FROM WORKER {} TO SERVER: {}\n".format(
                                        identity, reply_frame
                                    ),
                                    2,
                                )
                                self.print(
                                    "5. SENDING FROM WORKER {} TO SERVER: [client_address, task_key, task_success_signal, "
                                    "result]\n\n".format(identity),
                                    1,
                                )

                        except Exception as e:

                            task_key_bin = reply_frame_header[-1]
                            task_key = msgpack.unpackb(task_key_bin, raw=False)

                            warning_msg = "task id {} failed due to: {}".format(
                                task_key, repr(e)
                            )
                            warnings.warn(warning_msg, TASK_FAILED)

                            error_binary = msgpack.packb(repr(e), use_bin_type=True)

                            # [client_address, task_key, task_failure_signal, error]
                            reply_frame = reply_frame_header + [
                                TASK_FAILURE_SIGNAL,
                                error_binary,
                            ]

                            self.print(
                                "5. SENDING FROM WORKER {} TO SERVER: {}\n".format(
                                    identity, reply_frame
                                ),
                                1,
                            )
                            self.print(
                                "5. SENDING FROM WORKER {} TO SERVER: [client_address, task_key, task_failure_signal, "
                                "error]\n\n".format(identity),
                                1,
                            )

                        # The DEALER socket will prepend the worker address.
                        # [client_address, task_key, task_success_signal, result] or
                        # [client_address, task_key, numpy_task_success_signal, result] or
                        # [client_address, task_key, task_failure_signal, error]
                        worker.send_multipart(reply_frame)

                    else:

                        raise ValueError("Invalid message")

        except KeyboardInterrupt:

            print("Keyboard Interrupted")

            perform_final = True

        except ZMQError:

            perform_final = False

        except Exception as e:

            print("Worker Process Internal Failure: ", repr(e))

            reply_frame = frames[:2] + [TASK_FAILURE_SIGNAL] + task_payload

            worker.send_multipart(reply_frame)

            perform_final = True

        finally:

            self.print("GRACEFULLY TERMINATING WORKER {}".format(identity), 1)

            if perform_final:

                poller.unregister(worker)
                worker.setsockopt(zmq.LINGER, 0)
                worker.close()
                context.term()



import itertools
import logging
import os
import sys
import tempfile
import time
import warnings
from collections import defaultdict
from multiprocessing import set_start_method
from random import randint

import cloudpickle
import feather
import msgpack
import numpy as np
import pandas as pd
import psutil
import zmq

from .config import *
from .server import ServerProcess
from .utils import ProgressBar
from .worker import WorkerProcess


class WORKER_FAILED(UserWarning):
    pass


class Futures:
    """Parallelization using ZeroMQ message passing framework. The architecture
    design is based on the Paranoid Pirate Pattern described in
    "https://zguide.zeromq.org/docs/chapter4/".

    This class is responsible for starting the daemon process for the proxy server
    and the daemon processes for the workers. The client runs on the main process.
    When the client submits tasks, they are queued in the server proxy server socket.
    The server will load balance the tasks to the LRU (least recently used) worker.
    Once the worker finishes the task, it will send the results message to the server,
    which will in turn send the message back to the client in the main process.


                                                                      worker (client)
    client <--asynch msg queue--> proxy server <--asynch msg queue--> worker (client)
                                                                      worker (client)


    There are 5 different ways to submit the user functions:

    - submit
    - submit_keyed
    - submit_stateful
    - apply
    - capply

    Refer to the method docstrings below and the accompanying jupyter notebook
    tutorial for their use cases.

    Parameters
    ----------
    n_workers : int
        The number of worker processes to use. Default is the number of available
        physical cores.

    start_method : str, Default
        The process start method. Available options: "fork", "spawn", and "forkserver".
        Default is "fork".

    verbose : int
        The levels of verboseness. Available options: 0, 1, 2. Default is 0.

    worker : WorkerProcess
        The subclass of WorkerProcess for custom initialization of the worker process.

    worker_args : list
        The positional arguments for the WorkerProcess.

    worker_kwargs : dict
        The keyword arguments for the WorkerProcess.

    n_retries : int
        The number of retries for any given task.
    """

    def __init__(
        self,
        n_workers=None,
        start_method="fork",
        verbose=0,
        n_retries=2,
        worker=None,
        *worker_args,
        **worker_kwargs,
    ):

        self.n_workers = n_workers
        self.start_method = start_method
        self.verbose = verbose
        self.request_retries = n_retries
        self.worker = worker
        self.worker_args = worker_args
        self.worker_kwargs = worker_kwargs

        self.worker_procs = []

        if n_workers is None:

            self.n_workers = psutil.cpu_count(logical=False)

        set_start_method(self.start_method, force=True)

        self.dataframe = None
        self.results = dict()
        self.errors = dict()
        self._task_keys = set()
        self._pending_tasks = dict()
        self._failed_tasks = dict()
        self._fail_counter = defaultdict(int)

        self.server_client_online = False
        self.mode = "normal"
        self.sub_mode = defaultdict(bool)
        self.sub_mode_secondary_checker = dict()
        self.temp_dir = None

    def print(self, s, lvl):

        if self.verbose >= lvl:
            print(s)

    def start_workers(self, n_workers):
        """Start the worker processes, either the default WorkerProcess class
        or overridden WorkerProcess subclass.

        Parameters
        ----------
        n_workers : int
            The number of worker processes to use.
        """
        for _ in range(n_workers):

            if self.worker is not None:

                worker_proc = self.worker(
                    __verbose__=self.verbose,
                    __dataframe__=self.dataframe,
                    __forked__=self.start_method == "fork",
                    __mode__=self.mode,
                    __partition__=self.sub_mode["partition"],
                    *self.worker_args,
                    **self.worker_kwargs,
                )

            else:

                # If forked, passing ``self.dataframe`` won't create a copy.
                #
                worker_proc = WorkerProcess(
                    __verbose__=self.verbose,
                    __dataframe__=self.dataframe,
                    __forked__=self.start_method == "fork",
                    __mode__=self.mode,
                    __partition__=self.sub_mode["partition"],
                )

            worker_proc.daemon = True
            worker_proc.start()

            self.worker_procs.append(worker_proc)

    def start_server(self):
        """Start the server process."""
        self.server_process = ServerProcess(self.client_address, self.verbose)
        self.server_process.daemon = True
        self.server_process.start()

    def start_client(self):
        """Connect the client socket to the server endpoint. Assign the
        client the string address (the current process's PID).
        """
        self.client_address = str(os.getpid()).encode()
        self.context = zmq.Context()
        self.client = self.context.socket(zmq.DEALER)
        self.client.setsockopt(zmq.IDENTITY, self.client_address)
        self.client.connect(SERVER_ENDPOINT)

        self.print("CLIENT STARTED: {}\n\n".format(self.client_address), 1)

    def start(self):
        """Utility function to start the processes and bind the sockets.
        Also, the ``server_client_online`` flag is set to True, so that the
        Futures object knows not to restart the processes.
        """
        self.start_workers(self.n_workers)
        self.start_client()
        self.start_server()

        self.server_client_online = True

    def close(self):
        """Gracefully terminate all the network sockets and processes with
        timeout by sending them the kill signal. If there are remaining worker
        processes after the set timeout, those processes are forcefully
        terminated.
        """
        frames = [DUMMY_TASK_KEY, KILL_SIGNAL]

        # The DEALER socket will prepend the client address.
        # [dummy_task_key, kill_signal]
        self.client.send_multipart(frames)

        self.server_client_online = False

        close_start_time = time.time()

        while (
            any([proc.is_alive() for proc in self.worker_procs])
            and time.time() - close_start_time < 1
        ):

            time.sleep(0.01)

        for proc in self.worker_procs:

            if proc.is_alive():

                self.print("FORCEFULLY TERMINATING WORKER", 1)

                proc.terminate()

    def clear(self):
        """Clean out the computation results and data from memory."""
        self.dataframe = None
        self._task_keys = set()
        self._pending_tasks = dict()
        self.results = dict()
        self.errors = dict()
        self._failed_tasks = dict()
        self.sub_mode = defaultdict(bool)
        self.sub_mode_secondary_checker = dict()

    def apply_to(self, dataframe, groupby=None, orderby=None):
        """Prepare the dataframe for parallel processing. This procedure
        depends on the process start method.

        If the startmethod is forking, then we will rely on the Unix COW
        to bring the dataframe to the subprocesses. Otherwise, we will write
        the dataframe on disk using feather and read it from subprocesses.

        If groupby is not None, then we will partition the dataframe by the
        groupby column, and the dict will be saved in global namespace in
        case of forking, or partitioned dataframes on disk using feather
        otherwise (the feather writing of partitions is done in parallel using
        forking).

        Parameters
        ----------
        dataframe : Pandas dataframe

        groupby : str

        orderby : str

        """

        self.mode = "pandas"

        if groupby:

            self.sub_mode["partition"] = True

            dataframe_dict = dict()

            for key, sub_df in dataframe.groupby(groupby):

                if orderby:

                    sub_df = sub_df.sort_values(orderby)

                sub_df.reset_index(drop=True, inplace=True)
                sub_df = sub_df.copy(deep=False)

                dataframe_dict[key] = sub_df

            dataframe = dataframe_dict

        else:

            self.sub_mode["partition"] = False

            if orderby:

                dataframe = dataframe.sort_values(orderby)
                dataframe.reset_index(drop=True, inplace=True)
                dataframe = dataframe.copy(deep=False)

        if self.start_method == "fork":

            set_start_method(self.start_method, force=True)

            if groupby:

                self.dataframe = dataframe_dict

            else:

                self.dataframe = dataframe

        # If the processes are not forked, then write the dataframe on disk.
        # If groupby, then use forking to write the dataframe partitions in
        # parallel.
        else:

            assert self.start_method in ["spawn", "forkserver"]
            set_start_method(self.start_method, force=True)

            if self.sub_mode["partition"]:

                pass

            else:

                cur_dirpath = os.getcwd()
                temp_filepath = os.path.join(cur_dirpath, "__futures_zero_dataframe__")
                feather.write_dataframe(dataframe, temp_filepath)

                self.dataframe = temp_filepath

            # if groupby:

            #     global global_dataframe
            #     global_dataframe = dataframe

            #     def write_dataframe(key=None):

            #         global global_dataframe

            #         if isinstance(global_dataframe, dict):

            #             filename = "".join([TMP_FILENAME, str(key)])
            #             filepath = os.path.join(os.getcwd(), filename)

            #             feather.write_dataframe(global_dataframe[key], filepath)

            #         else:

            #             print(global_dataframe)

            #     for key in dataframe.keys():

            #         self.submit_keyed(key, write_dataframe, key)

    def apply(self, func, *args, __key__=None, **kwargs):
        """Apply the user function on the Pandas dataframe passed in the ``apply_to``
        method. The ``func`` method signature requires the dataframe as the first
        positional argument.

        Parameters
        ----------
        func : Python method

        __key__ : Python object
        """
        self.mode = "pandas"

        key = __key__ or len(self._task_keys)

        if isinstance(key, list):
            key = tuple(key)

        binary = self._serialize(key, False, func, *args, **kwargs)

        self._submit(key, binary)

    def capply(self, column=None, func=None, *args, **kwargs):
        """``capply`` stands for column-apply. Same as the ``apply`` method,
        but the ``func`` method has an additional requirement that the return
        value is the numpy array with the same length as the dataframe. That
        array will be appended to the input dataframe with the name ``column``.

        You can also give the ``column`` a list of string names if the return value
        of the ``func`` is a multidimensional array. Then the dataframe will get
        multiple columns, using column names from the ``column`` list.

        Parameters
        ----------
        func : Python method

        column : str or list or str

        args : Python objects
            Positional arguments

        """
        self.sub_mode["column"] = True

        key = column or len(self._task_keys)

        if isinstance(key, list):
            key = tuple(key)

        self.sub_mode_secondary_checker[key] = "column"

        self.apply(func, *args, __key__=key, **kwargs)

    def submit(self, func, *args, __key__=None, __stateful__=False, **kwargs):
        """Pass in user func to compute in parallel.

        Parameters
        ----------
        func : Python method

        __key__ : Python object

        __stateful__ : boolean
            If True, the ``func`` method must have "self" as the first positional
            argument, where that self is the WorkerProcess instance.
        """
        self.mode = "normal"

        key = __key__ or len(self._task_keys)

        if isinstance(key, list):
            key = tuple(key)

        binary = self._serialize(key, __stateful__, func, *args, **kwargs)

        self._submit(key, binary)

    def submit_keyed(self, key, func, *args, __stateful__=False, **kwargs):
        """Same as ``submit``, but key positional argument is required.

        Parameters
        ----------
        key : Python object

        func : Python method
        """
        self.submit(func, *args, __key__=key, __stateful__=__stateful__, **kwargs)

    def submit_stateful(self, func, *args, __key__=None, **kwargs):
        """Same as ``submit`` but the user method signature has a requirement
        that "self" is the first positional argument. This "self" is the
        WorkerProcess instance. This method is same as ``submit`` with
        __stateful__ set to True.

        Parameters
        ----------
        func : Python method

        __key__ : Python object
        """
        self.submit(func, *args, __key__=__key__, __stateful__=True, **kwargs)

    def submit_stateful_keyed(self, key, func, *args, **kwargs):
        """Utility function combining ``submit_keyed`` and ``submit_stateful``.

        Parameters
        ----------
        func : Python method

        key : Python object
        """
        self.submit(func, *args, __key__=key, __stateful__=True, **kwargs)

    def _serialize(self, key, stateful, func, *args, **kwargs):
        """The ``func`` method is serialized using cloudpickle. The arguments
        are serialized using msgpack. We will also attach 3 signals to the
        message: task mode (e.g. normal, pandas), process start method
        (e.g. fork), and statefulenss message.

        The return value is the message that will be passed over to the server,
        which will route the message in turn to the LRU workers. This message
        consists of a list of binary streams:

        0. key
        1. task mode
        2. process start method
        3. statefulness of method
        4. func
        5. args (binary of args and kwargs)
        """

        # Serialize the task payload, comprised of the user method and method arguments.
        binary_func = cloudpickle.dumps(func)
        binary_args = msgpack.packb([args, kwargs], use_bin_type=True)
        binary_key = msgpack.packb(key, use_bin_type=True)

        task_ctrl_msg = []

        # Task mode message.
        if self.mode == "normal":

            task_ctrl_msg.append(NORMAL_TASK_REQUEST_SIGNAL)  # b"\x04"

        elif self.mode == "pandas":

            if self.sub_mode["partition"]:

                task_ctrl_msg.append(PANDAS_PARTITION_TASK_REQUEST_SIGNAL)  # b"\x05"

            else:

                task_ctrl_msg.append(PANDAS_NONPARTITION_TASK_REQUEST_SIGNAL)  # b"\x06"

        # Process start method message.
        if self.start_method == "fork":

            task_ctrl_msg.append(FORKED_PROCESS_SIGNAL)  # b"\x07"

        elif self.start_method in ["spawn", "forkserver"]:

            task_ctrl_msg.append(SPAWNED_PROCESS_SIGNAL)  # b"\x08"

        else:

            raise ValueError("Not a valid start method.")

        # User method statefulness message.
        if stateful:

            task_ctrl_msg.append(STATEFUL_METHOD_SIGNAL)  # b"\x10"

        else:

            task_ctrl_msg.append(STATELESS_METHOD_SIGNAL)  # b"\x11"

        return [binary_key] + task_ctrl_msg + [binary_func, binary_args]

    def _submit(self, key, request):
        """Utility function to send work request to the server (which will in turn delegate
        the message to the first available LRU worker). Once the work is submitted, that
        request message will be added to pending tasks list and the key will be added to the
        task keys list. Importantly, if the server client are not online, we will call the
        utility start method to start the daemon processes.

        Parameters
        ----------
        key : binary

        request : list of binaries
                        0. key
                        1. task mode
                        2. process start method
                        3. statefulness of method
                        4. func
                        5. args (binary of args and kwargs)
        """
        if not self.server_client_online:
            self.start()
            time.sleep(0.05)

        self.print(f"1. SENDING FROM CLIENT TO SERVER: {request}\n", 2)
        self.print("** NOTE: b'\\x92\\x90\\x80' is [[], {}]\n\n", 2)
        self.print(
            "1. SENDING FROM CLIENT TO SERVER: [task_key, task_mode_signal, start_method_signal, "
            "func_statefulness_signal, func, args]",
            1,
        )

        # The DEALER socket will prepend the client address.
        # [task_key, task_mode_signal, start_method_signal, func_statefulness_signal, func, args]
        self.client.send_multipart(request)

        self._pending_tasks[key] = request
        self._task_keys.add(key)

    def _handle_failed_tasks(self, task_key, error_msg):
        """Failed task retry mechanism. If the task has failed less than the set threshold counts,
        re-submit the task. Otherwise, remove the task from pending tasks list and put it in the
        failed tasks list and record the error.
        """
        self._fail_counter[task_key] += 1

        if self._fail_counter[task_key] < self.request_retries + 1:

            self.print(
                f"9. TASK {task_key} FAILED {self._fail_counter[task_key]} TIME(S), RETRYING\n\n",
                1,
            )

            self._submit(task_key, self._pending_tasks.pop(task_key, None))

        else:

            self.print(
                f"9. TASK {task_key} FAILED {self._fail_counter[task_key]} TIME(S), ABORTING\n\n",
                1,
            )

            # Place the task in the failed tasks dictionary.
            self._failed_tasks[task_key] = self._pending_tasks.pop(task_key, None)

            self.errors[task_key] = error_msg

    def _poll(self):
        """Start the infinite while loop to listen to the server for replies. Continue the
        loop while the sum of results, errors and dead workers is less than the total number
        of submitted tasks.

        There are two types of successful signals:
            - normal: the payload is deserialized using msgpack.
            - numpy: the numpy array is reconstructed using memoryview from the memory buffer

        There are two types of failure signals:
            - task failure: error caused by the user ``func``.
            - worker failure: error caused by death of the process itself.

        The server cannot see the worker death and therefore the client from main process
        will check if there are workers that are dead. If there are, it will start as many
        processes as they died. The tasks that were never completed by those processes will
        be re-submited.
        """
        while len(self.results) + len(self.errors) < len(self._task_keys):

            # Start listening to replies from the server
            if (self.client.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:

                # The REQ socket stripped off the client address.
                # [task_key, task_signal, error_msg, task_signal, task_signal, func, args] or
                # [task_key, task_signal, result] or
                # [dummy_task_key, worker_failure_signal, failed_task_keys]
                reply = self.client.recv_multipart()

                task_key = msgpack.unpackb(reply[0], raw=False)  # task_key

                if isinstance(task_key, list):
                    task_key = tuple(task_key)

                reply_payload = reply[1:]
                # [task_success_signal, result]
                # [numpy_task_success_signal, metadata, result]
                # [task_signal, error_msg]
                # [worker_failure_signal, failed_task_keys]

                # If the task is returning general python object(s), use msgpack to deserialize data.
                if reply_payload[0] == TASK_SUCCESS_SIGNAL:

                    self.print(f"8. RECEIVED FROM SERVER IN CLIENT: {reply}\n", 2)
                    self.print(
                        "8. RECEIVED FROM SERVER IN CLIENT: [task_key, task_success_signal, result]\n\n",
                        1,
                    )

                    # Recover the returned data.
                    result = msgpack.unpackb(reply_payload[1], raw=False)

                    self.bar.report(len(self.results))

                    # Remove the task from pending status only after checking the results.
                    self._pending_tasks.pop(task_key, None)

                    self.results[task_key] = result

                # If the task returns a single numpy array, use memoryview and memory buffer to recover the data.
                elif reply_payload[0] == NUMPY_TASK_SUCCESS_SIGNAL:

                    self.print(f"8. RECEIVED IN CLIENT FROM SERVER: {reply}\n", 2)
                    self.print(
                        "8. RECEIVED IN CLIENT FROM SERVER: [task_key, numpy_task_success_signal, metadata, result]\n\n",
                        1,
                    )

                    # Recover the numpy array from the binary stream.
                    metadata = msgpack.unpackb(reply_payload[1], raw=False)
                    buf = memoryview(reply_payload[2])
                    result = np.frombuffer(buf, dtype=metadata["dtype"]).reshape(
                        metadata["shape"]
                    )

                    if self.mode == "pandas":

                        # If ``capply`` was called, append the array as a column to the input dataframe.
                        if (
                            self.sub_mode["column"]
                            and self.sub_mode_secondary_checker.get(task_key)
                            == "column"
                        ):

                            # If groupby was not used
                            if not self.sub_mode["partition"]:

                                self.print(
                                    "9. MODE: PANDAS, SUBMODES: COLUMN\n\n",
                                    1,
                                )

                                if (
                                    (isinstance(result, np.ndarray))
                                    & (len(result) == len(self.dataframe))
                                    & (not isinstance(task_key, int))
                                ):

                                    # If the ``column`` was made of multiple column names.
                                    if isinstance(task_key, tuple):

                                        # Only if the number of returned columns equals the suggested
                                        # number of columns.
                                        if result.shape[1] == len(task_key):

                                            res_df = pd.DataFrame(
                                                result, columns=task_key
                                            )
                                            self.dataframe = self.dataframe.join(res_df)
                                            self.results[task_key] = None

                                        else:

                                            warning_msg = (
                                                "The resulting column could not be attached to "
                                                "the dataframe because result has different number of columns. "
                                                "Recording the results in 'results' dict"
                                            )
                                            warnings.warn(warning_msg)

                                            self.results[task_key] = result

                                    # If the ``column`` was a single column name.
                                    else:

                                        # Only if the returned array is 1-dimensional.
                                        if len(result.shape) == 1:

                                            self.dataframe[task_key] = result
                                            self.results[task_key] = None

                                        else:

                                            warning_msg = (
                                                "The resulting column could not be attached to "
                                                "the dataframe because result has too many columns. "
                                                "Recording the results in 'results' dict"
                                            )
                                            warnings.warn(warning_msg)

                                            self.results[task_key] = result

                                # Either the result is not ndarray or
                                # the length of the returned array is not the same as the dataframe's or
                                # task_key is an integer value.
                                else:

                                    print(result, "+++")
                                    print(task_key, "+++")

                                    warning_msg = (
                                        "The resulting column could not be attached to "
                                        "the dataframe. Recording the results in 'results' dict"
                                    )
                                    warnings.warn(warning_msg)

                                    self.results[task_key] = result

                            # If groupby was used
                            else:

                                self.print(
                                    "9. MODE: PANDAS, SUBMODES: COLUMN, PARTITION\n\n",
                                    1,
                                )

                                raise NotImplementedError(
                                    "Parition submode is not yet supported"
                                )

                        # If only ``apply`` was called and not ``capply``.
                        else:

                            self.print(
                                "9. MODE: PANDAS, SUBMODES:\n\n",
                                1,
                            )

                            self.results[task_key] = result

                    elif self.mode == "normal":

                        self.print(
                            "9. MODE: NORMAL, SUBMODES:\n\n",
                            1,
                        )

                        self.results[task_key] = result

                    self.bar.report(len(self.results))

                    # Remove the task from pending status.
                    self._pending_tasks.pop(task_key, None)

                elif reply_payload[0] == TASK_FAILURE_SIGNAL:

                    self.print(f"8. RECEIVED IN CLIENT FROM SERVER: {reply}\n", 2)
                    self.print(
                        "8. RECEIVED IN CLIENT FROM SERVER: [task_key, task_failure_signal, error]\n\n",
                        1,
                    )

                    error_msg = msgpack.unpackb(reply_payload[1], raw=False)

                    self._handle_failed_tasks(task_key, error_msg)

                # If worker failure message was sent from this client to the server (refer below)
                # 1. restart the number of failed worker processes
                # 2. retry the unfinished tasks as needed.
                elif reply_payload[0] == WORKER_FAILURE_SIGNAL:

                    self.print(f"8. RECEIVED IN CLIENT FROM SERVER: {reply}\n", 2)
                    self.print(
                        "8. RECEIVED IN CLIENT FROM SERVER: [dummy_task_key, worker_failure_signal, error]\n\n",
                        1,
                    )

                    failed_task_keys = msgpack.unpackb(reply_payload[1], raw=False)

                    for failed_task_key in failed_task_keys:

                        warning_msg = "task id {} failed due to: worker death".format(
                            failed_task_key
                        )
                        warnings.warn(warning_msg, WORKER_FAILED)

                        self._handle_failed_tasks(
                            failed_task_key, "Premature worker death"
                        )

            # Check the status of the child processes every REQUEST_TIMEOUT milliseconds.
            worker_deaths = [not proc.is_alive() for proc in self.worker_procs]

            # If any workers are dead, inform the server and start that many workers again.
            if any(worker_deaths):

                # Update the worker process list.
                self.worker_procs = [
                    proc for proc in self.worker_procs if proc.is_alive()
                ]

                frames = [DUMMY_TASK_KEY, WORKER_FAILURE_SIGNAL]

                # The DEALER socket will prepend the client address.
                # [dummy_task_key, worker_failure_signal]
                self.client.send_multipart(frames)

                self.start_workers(sum(worker_deaths))

    def get(self, keyed=False):

        self.bar = ProgressBar()
        self.bar.set_total(len(self._task_keys))
        self.bar.report(0)

        self._poll()

        if len(self.results) == len(self._task_keys):

            self.bar.completion_report()

        self.close()

        if self.mode == "normal":

            if not keyed:
                return list(self.results.values())
            else:
                return self.results

        elif self.mode == "pandas":

            return self.dataframe



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

import setuptools
from setuptools.command.install import install

# circleci.py version
from src import futures_zero

VERSION = f"v{futures_zero.__version__}"


def readme():
    """print long description"""
    with open("README.rst") as f:
        return f.read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="futures-zero",
    version=VERSION,
    author="Jay Kim",
    author_email="mozjay0619@gmail.com",
    description="Parallelization tool with the futures API using ZeroMQ Paranoid Pirate Pattern.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/mozjay0619/futures-zero",
    license="DSB 3-clause",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3",
    cmdclass={
        "verify": VerifyVersionCommand,
    },
)
